// Jib Goes Shopping
// (C) 2006, 2007 Ulrich Hecht (ulrich.hecht@gmail.com)
// Licensed under the terms of the GNU Public License v2

// Compile with:
// zcc +zx -vn -O3 game.c -o jibshop.bin -create-app -lim2 -lsp1 -lballoc -lrect

// character codes for bg tiles
#define TILE_HAUS 200
#define TILE_PFLASTER 210

#include <sprites/sp1.h>
#include <balloc.h>
#include <spectrum.h>
#include <games.h>
#include <im2.h>
#include <input.h>

#define myabs(a) ((a)>0?(a):-(a))

/*******************LIBRARY DECLARATIONS********************/

//// SP1.LIB ////

uchar idtypetbl[] = {                            // (id,type) pairs
   SP1_ID_MASK2   , 0,
   SP1_ID_MASK2_LB, 1,
   SP1_ID_MASK2_RB, 2,
   0              , 0                            // always end table in (0,0)
};

extern uchar sprdrawftbl[];                      // must be declared in ASM to access ASM names

#asm

._sprdrawftbl                                    ; table of sprite draw functions indexed by type
   defw SP1Mask2Data
   defw SP1Mask2LBData
   defw SP1Mask2RBData
   
LIB SP1Mask2Data, SP1Mask2LBData, SP1Mask2RBData ; get functions from sp1 library

XDEF _im2_vtable
DEFC _im2_vtable = $d300

#endasm

/* volatile */ uchar frames = 0;

// interrupt handler
void my_isr(void)
{
  frames++;
  return_nc;	// return with carry clear
}

void my_printat(uchar row, uchar col, uchar color, uchar* string)
{
  uchar c = col;
  struct sp1_Rect rt;
  rt.col = col;
  rt.row = row;
  rt.height = 1;
  while(*string) {
    sp1_PrintAt(row, c++, color, *string++);
  }
  rt.width = c - col;
  sp1_Invalidate(&rt);
}

// print a number to a string; leaner than sprintf
void my_snprint(uchar* string, int num, uchar len)
{
  int pos = len - 2;
  uchar i;
  for(i=0; i<len-1; i++) string[i]=' ';
  string[len-1]=0;
  string[pos--]=string[pos--]='0';
  while(num) {
    string[pos--] = num % 10 + '0';
    num /= 10;
  }
  string[pos]='-';
}

// callback for SP1 library
void colorSpr(uint count, struct sp1_cs* c)
{
  c->attr_mask = 0;
  c->attr = INK_BLUE|PAPER_WHITE;
}

char joytype = -1;	// joystick interface type
uint keycodes[4];	// scancodes for game keys

// Clipping Rectangle for Sprites
struct sp1_Rect cr = {0, 0, 32, 24};             // full screen

extern uint rand(void);
int int_seed;	// seed for rand()


//// BALLOC.LIB ////

BAQTBL(2)                                        // two memory queues

#pragma output STACKPTR=54016                    // move stack to $d300, needed by sp1.lib


// Memory allocation functions for sp1.lib

void *u_malloc(uint size) {
    if (size==sizeof(struct sp1_ss))             // allocate memory for sprite structs from queue 0
       return ba_Malloc(0);
    return ba_Malloc(1);                         // allocate memory for char structs from queue 1
}

void u_free(void *addr) {
    ba_Free(addr);
}


// Sprite data structure

struct sprentry {
   struct sp1_ss  *s;                            // sprite handle returned by sp1_CreateSpr()
#ifdef USE_EVEN
   char           dx_even;                            // signed horizontal speed in pixels
   char           dy_even;                            // signed vertical speed in pixels
#endif
   char           dx_odd;                            // signed horizontal speed in pixels
   char           dy_odd;                            // signed vertical speed in pixels
   uchar animcount;	// counter used for animation
   uchar destx;		// destination for uli NPC (x coord)
   uchar desty;		// destination for uli NPC (y coord)
   uchar lastdir;	// previous direction uli was moving in
   uint offtime;	// time for prizes to stay invisible
   uint ontime;		// time for prizes to stay visible
};

struct sp1_Rect mazerect = { 2, 1, 30, 20 };	// maze coordinates on screen
struct sp1_tp* maze;	// the maze itself, put on heap to avoid bloating the binary
//struct sp1_tp maze[20*30];

uchar lives;			// player lives
uint score;			// player score
uint hiscore;			// hiscore
struct sprentry uli;		// NPC
struct sprentry jib;		// player character
struct sprentry prizes[3];	// prizes (diamonds)

// Attach C Variables to Sprite Graphics Declared in ASM at End of File

extern uchar ulifront1[];
extern uchar ulifront15[];
extern uchar ulifront2[];
extern uchar ulifront25[];
extern uchar uliback1[];
extern uchar uliback2[];
extern uchar ulileft1[];
extern uchar ulileft2[];
extern uchar uliright1[];
extern uchar uliright2[];

extern uchar jibfront1[];
extern uchar jibfront2[];
extern uchar jibback1[];
extern uchar jibback2[];
extern uchar jibleft1[];
extern uchar jibleft2[];
extern uchar jibright1[];
extern uchar jibright2[];

extern uchar haus[];
extern uchar pflaster[];

extern uchar diamond1[];
extern uchar diamond15[];

// intro screen, joystick interface selection
void welcome()
{
  uchar i,k;
  uchar* frame;
  uchar num[2];
  num[1] = 0;
  sp1_ClearRect(&cr, INK_BLACK|PAPER_YELLOW,' ',0xff);
  sp1_Invalidate(&cr);
  my_printat(4,7,INK_BLUE|PAPER_YELLOW,"Jib Goes Shopping");
  my_printat(7,5,INK_RED|PAPER_YELLOW,"Happy Birthday, Jib!");
  sp1_MoveSprAbs(jib.s, &cr, jibleft1, 10, 15, 0, 0);
  sp1_MoveSprAbs(prizes[0].s, &cr, diamond1, 10, 10, 0, 0);
  sp1_MoveSprAbs(prizes[1].s, &cr, diamond1, 10, 20, 0, 0);
  my_printat(15,5,INK_BLACK|PAPER_YELLOW,"Press a key to start.");
  my_printat(22,2,INK_BLACK|PAPER_YELLOW,"(C) 2006, 2007 Ulrich Hecht");
  my_printat(23,4,INK_BLACK|PAPER_YELLOW,"ulrich.hecht@gmail.com");
  sp1_UpdateNow();
  while(in_Inkey() == 0) {
    int_seed++; 
    if((int_seed & 4095) == 0) {
      if(int_seed & 4096) frame = jibright1;
      else frame = jibleft1;
      sp1_MoveSprAbs(jib.s, &cr, frame, 10, 15, 0, 0);
      sp1_UpdateNow();
    }
  }
  
  if(joytype == -1) {	// only ask for the joystick interface if it has not been set before
    sp1_ClearRect(&cr, INK_BLACK|PAPER_YELLOW, ' ', 0xff);
    sp1_Invalidate(&cr);
    my_printat(3, 2, INK_BLACK|PAPER_YELLOW, "Pick your joystick interface");
    
    // show all available joystick interfaces...
    for(i = 1; i < GAME_DEVICES + 1; i++) {
      num[0] = '0' + i;
      my_printat(5 + i * 2, 5, INK_BLACK|PAPER_YELLOW, num);
      my_printat(5 + i * 2, 7, INK_BLACK|PAPER_YELLOW, joystick_type[i-1]);
    }
    // ...and keyboard
    num[0] = '0' + i;
    my_printat(5 + i * 2, 5, INK_BLACK|PAPER_YELLOW, num);
    my_printat(5 + i * 2, 7, INK_BLACK|PAPER_YELLOW, "Keyboard (Q/A/O/P)");

    sp1_UpdateNow();
    while(joytype == -1) {
      k = in_Inkey();
      if(!k) continue;	// no key pressed
      k -= '0';	// ASCII -> int
      if(k > 0 && k <= GAME_DEVICES) joytype = k;
      if(k == GAME_DEVICES + 1) {
        joytype = -2; // keyboard
        keycodes[0] = in_LookupKey('q');
        keycodes[1] = in_LookupKey('a');
        keycodes[2] = in_LookupKey('o');
        keycodes[3] = in_LookupKey('p');
      }
    }
  }
    
#ifdef DEBUG
  sp1_PrintAt(0,0,INK_BLACK|PAPER_WHITE,'0'+joytype);
#endif
  sp1_Invalidate(&cr);
}

// check if sprite eins overlaps with sprite zwei. Assumes sprite size of 16x16.
int collide(struct sp1_ss* eins, struct sp1_ss* zwei)
{
  struct r_Rect8 r1, r2;
  sp1_MakeRect8Pix(eins, &r1);
  sp1_MakeRect8Pix(zwei, &r2);
  r1.width=r2.width=r1.height=r2.height=16;
  return r_IsRectInRect8(&r1, &r2);
}

// place a tile in the maze (a maze tile is equivalent to 2x2 screen tiles)
void settile(uchar tile, uchar color, uchar x, uchar y)
{
  uint pos=x*2+y*15*2*2;	// the compiler likes "15*2*2" (shift) better than "30*2" (mult)
  maze[pos].attr=maze[pos+1].attr=maze[pos+30].attr=maze[pos+31].attr= color;
  maze[pos].tile=tile; maze[pos+1].tile=tile+1;
  maze[pos+30].tile=tile+2; maze[pos+31].tile=tile+3;
}

// retrieve the maze tile (more precisely, the top left screen tile of the given maze tile) number
uchar gettile(uchar x, uchar y)
{
  return maze[x * 2 + y * 15 * 2 * 2].tile;	// see settile() for rationale behind "15*2*2".
}

// print the life bar
void printlives(void)
{
  struct sp1_tp lifebar[4];	// one tile for each life
  static struct sp1_Rect lifebarrect = {1, 1, 4, 1};	// screen position of life bar
  uchar i;

#if 1	// for debugging: set to 0 to print a numerical value instead
  // print lives number of 'J's.
  for(i=0; i != 4; i++)
  {
    lifebar[i].attr = INK_MAGENTA|PAPER_WHITE;
    if(i<lives)
      lifebar[i].tile = 'J';
    else
      lifebar[i].tile = ' ';
  }
#else
  lifebar[0].attr = INK_MAGENTA|PAPER_WHITE;
  lifebar[0].tile = '0'+lives;
#endif
  sp1_PutTiles(&lifebarrect, lifebar);
  sp1_Invalidate(&lifebarrect);
}

// move player, NPC, and prizes to their initial position:
// player: bottom left corner
// NPC: top right corner
// prizes: off screen
void resetsprites(void)
{
   uchar i;
   
   // move NPC to his corner and make it stay there, reset animation counter,
   // set speed to 1 on even frames, 2 on odd frames (even not used as of now)
   sp1_MoveSprAbs(uli.s, &cr, ulifront1, 4, 27, 0, 0);
   uli.destx=27;
   uli.desty=4;
   uli.animcount=0;
#ifdef USE_EVEN
   uli.dx_even=uli.dy_even=1;
#endif
   uli.dx_odd=uli.dy_odd=2;
   
   // likewise for player
   sp1_MoveSprAbs(jib.s, &cr, jibback1, 18, 3, 0, 0);
   jib.animcount=0;
#ifdef USE_EVEN
   jib.dx_even=jib.dy_even=2;
#endif
   jib.dx_odd=jib.dy_odd=2;
   
   for(i=0 ; i!=3 ; i++)
   {
     sp1_MoveSprAbs(prizes[i].s, &cr, diamond1, 0, 32, 0, 0);	// move prizes off screen
     // set prizes to off for a random amount of time.
     prizes[i].ontime=0;
     prizes[i].offtime=rand()%512;
   }
   printlives();
}

// called when the player gets caught by the NPC
void lose_life(void)
{
  lives--;

  // do something funny
  my_printat(11,12,INK_MAGENTA|PAPER_WHITE,"GOT YOU!");
  sp1_UpdateNow();
  frames=0;
  while(frames<60*1) {}	// wait for one second (assuming 60 frames per second, which,
                        // thinking about it, is pretty stupid because the Spectrum
                        // is PAL-only... :)

  // redraw the maze, reset the game objects
  sp1_PutTiles(&mazerect, maze);
  sp1_Invalidate(&mazerect);
  resetsprites();
}

// called when the player lost her last life.
void game_over(void)
{
  // do something funny
  my_printat(11,11,INK_RED|PAPER_BLACK,"GAME OVER");
  
  if(score > hiscore) {
    my_printat(13,9,INK_YELLOW|PAPER_BLACK,"NEW HI-SCORE!");
    hiscore = score;
  }
  
  sp1_UpdateNow();

  frames=0;
  while(frames<60*3) {}	// wait for 3 NTSC seconds (see lose_life())
}

// speed toggle. the idea was that characters move at different speeds
// depending on whether the frame number is even or odd, allowing more
// fine-grained control of the character speeds. however, using this
// feature introduces funny bugs with the NPC, such as getting stuck
// in dead ends and developing a tendency to go to the upper left corner.
// as it turns out, setting the player and the NPC to the same speed is
// just perfect for gameplay, so i didn't bother to fix it.
#ifdef USE_EVEN
uchar even=0;
#endif

// called every frame, takes care that the NPC reaches its current target
uchar moveuli(void)
{
  struct r_Rect8 r;	// place to save the current NPC position
  uchar destxpix, destypix;	// destination coords of the NPC in pixels
  uchar dx=0;		// x speed
  uchar dy=0;		// y speed
  uchar* anim;		// pointer to animation frame
  
  sp1_MakeRect8Pix(uli.s,&r);	// get the NPCs position
  // convert the NPC's destination to pixels
  destxpix=uli.destx * 8;	
  destypix=uli.desty * 8;
  
  // have we reached our destination already?
  if(r.x == destxpix && r.y == destypix) return 1;
  
#ifdef USE_EVEN
  if(myabs(r.x-destxpix) == 1 || myabs(r.y-destypix) == 1) {
    // the idea was to make sure that if the NPC is only 1 pixel away from
    // its destination, he should only move in 1 pixel steps to avoid
    // running beyond his destination, turning back, running over it again,
    // turning back, etc. i don't remember if it worked, though.
    dx = 1; dy = 1;
  } else {
    // otherwise, the walking speed depends on the frame odd- or evenness.
    if(even) {
      dx = uli.dx_even;
      dy = uli.dy_even;
    } else {
      dx = uli.dx_odd;
      dy = uli.dy_odd;
    }
  }
#else
  // simpler version if even is always 0
  dx = uli.dx_odd;
  dy = uli.dy_odd;
#endif
  
  // check what direction we have to move to and set the deltas accordingly.
  // (NB: ulisnextstep() takes care of making sure that we can actually move there.)
  if(r.x < destxpix) { anim = uliright1; dx = dx; dy = 0; uli.lastdir = MOVE_RIGHT; }
  if(r.x > destxpix) { anim = ulileft1; dx = -dx; dy = 0; uli.lastdir = MOVE_LEFT; }
  if(r.y < destypix) { anim = ulifront1; dx = 0; dy = dy; uli.lastdir = MOVE_DOWN; }
  if(r.y > destypix) { anim = uliback1; dx = 0; dy = -dy; uli.lastdir = MOVE_UP; }
  
  uli.animcount++;
  // move the sprite and set one of the 2 animation phases depending on animcount. 
  sp1_MoveSprRel(uli.s, &cr, (uli.animcount&8) ? anim : anim + (ulifront2 - ulifront1), 0, 0, dy, dx);
  return 0;
}

// decides where the NPC should move next.
void ulisnextstep(void)
{
  uchar ulimazex;	// NPC's position in the maze (x coord)
  uchar ulimazey;	// NPC's position in the maze (y coord)
  uchar newdir = 0;	// new direction to move to
  uchar notgoback;	// whether it is appropriate to go back where we came from
  struct r_Rect8 r;	// the player's position
  
  // convert sp1.lib coordinates to maze coordinates
  ulimazex = (uli.destx - 1) / 2;
  ulimazey = (uli.desty - 2) / 2;
  
  // check if I can see jib
  // this code cheats a little by ignoring walls in its line of sight for the sake
  // of simplicity. playtesting results suggest that players do not notice this
  // little shortcut, and it does in fact not affect gameplay adversely.
  sp1_MakeRect8Pix(jib.s,&r);	// get her pos. in pixels
  r.x >>= 3; r.y >>= 3;		// convert from pixels to 8x8 characters
  r.x -= 1; r.y -=2;		// substract the position of the maze on screen
  r.x >>= 1; r.y >>= 1;		// convert from characters to maze tiles (2x2 chars)
  
  // if we are one the same horizontal line as the player, move towards the player if
  // possible (i.e. if there is no wall in front of us).
  if(ulimazex == r.x) {
    if(ulimazey < r.y && gettile(ulimazex, ulimazey+1) == TILE_PFLASTER) newdir = MOVE_DOWN;
    else if(ulimazey > r.y && gettile(ulimazex, ulimazey-1) == TILE_PFLASTER) newdir = MOVE_UP;
#ifdef USE_EVEN
    // speed up
    uli.dy_even=uli.dx_even=2;
#endif
  }
  else if(ulimazey == r.y) { // same for the vertical line
    if(ulimazex < r.x && gettile(ulimazex+1, ulimazey) == TILE_PFLASTER) newdir = MOVE_RIGHT;
    else if(ulimazex > r.x && gettile(ulimazex-1, ulimazey) == TILE_PFLASTER) newdir = MOVE_LEFT;
#ifdef USE_EVEN
    // speed up
    uli.dx_even=uli.dy_even=2;
#endif
  }
  
#ifdef USE_EVEN
  // if we cannot see the player (no new direction is set at this point),
  // slow down to normal speed.
  if(!newdir) uli.dx_even=uli.dy_even=1;
#endif

#if 1	// set to 0 to disable NPC movement except when it can see the player.
  // figure out a new direction to walk to
  while(!newdir) {
    // pick a direction at random.
    // we are recycling the direction definitions from games.h
    newdir = 1 << (rand() % 4);    // MOVE_UP, MOVE_DOWN, ...
    
    // try not to walk back in most cases. allow walking back with
    // a low probability to avoid getting stuck in dead ends.
    // it would have been possible to check if we are in a dead end
    // and thus have no other chance but to walk back, but this would
    // need a lot of code, and i was not in the mood to write it.
    notgoback = rand() % 8;
    
    // check if we can actually move into that direction. if not, try again.
    // also check if walking into that direction would mean to walk back
    // where we came from, and only do so if notgoback allows us to.
    switch(newdir) {
      case MOVE_LEFT:
        if((uli.lastdir == MOVE_RIGHT && notgoback) ||
           gettile(ulimazex-1,ulimazey) == TILE_HAUS) { newdir = 0; continue; }
        break;
      case MOVE_RIGHT:
        if((uli.lastdir == MOVE_LEFT && notgoback) ||
           gettile(ulimazex+1,ulimazey) == TILE_HAUS) { newdir = 0; continue; }
        break;
      case MOVE_UP:
        if((uli.lastdir == MOVE_DOWN && notgoback) ||
           gettile(ulimazex,ulimazey-1) == TILE_HAUS) { newdir = 0; continue; }
        break;
      case MOVE_DOWN:
        if((uli.lastdir == MOVE_UP && notgoback) ||
           gettile(ulimazex,ulimazey+1) == TILE_HAUS) { newdir = 0; continue; }
        break;
    }
  }
#endif
  
  // figure out the new NPC destination and convert it back to sp1.lib coords
  switch(newdir)
  {
    case MOVE_LEFT:
      ulimazex--; break;
    case MOVE_RIGHT:
      ulimazex++; break;
    case MOVE_UP:
      ulimazey--; break;
    case MOVE_DOWN:
      ulimazey++; break;
  }
  uli.destx = ulimazex * 2 + 1;
  uli.desty = ulimazey * 2 + 2;
  
  moveuli();
}

// check if the given rectangle in the maze contains a wall.
// not reliable if the rectangle is larger than 2x2 chars, but for
// our sprites, that's just fine. uses character (sp1.lib) coordinates
uchar checkforhouse(uchar x1, uchar x2, uchar y1, uchar y2)
{
  return (gettile(x1, y1) == TILE_HAUS ||
         gettile(x1, y2) == TILE_HAUS ||
         gettile(x2, y1) == TILE_HAUS ||
         gettile(x2, y2) == TILE_HAUS);
}

// converts pixel coordinates to maze coordinates.
// x: pixel coordinate
// offset: offset of the maze from the top/left screen edge (characters)
uchar pixtomaze(uchar x, uchar offset)
{
  return ((x >> 3) - offset) >> 1;
}

// move a sprite by a pixel offset and set the appropriate animation frame
// sprite: pointer to sprite structure
// dx,dy: pixel offsets
// anim: pointer to two animation frames.
// this function assumes that all sprites are the same size and have two animation phases. (it uses
// the difference between the two jib animation frames to get the distance between two frames.)
// the raison d'etre for this function was to cut down on code size. now, it is only ever used once,
// so it would probably be a good idea to 
//void mymove(struct sprentry* sprite, uchar dy, uchar dx, uchar* anim)
//{
//}

// try to move the sprite by a given offset. before doing so, this function checks
// if there are walls that prevent this. if there are, the function tries to be
// as clever as possible in guessing what the player wants to be done.
void trymove(struct sprentry* sprite, uchar dy, uchar dx, uchar* anim)
{
  struct r_Rect8 r;	// holds the current sprite position
  uchar x1,x2,y1,y2;	// edges of the sprite after moving
  uchar yy1,yy2;	// modified y bounds used when trying an alternative move
  uchar xx1,xx2;	// modified x bounds used when trying an alternative move
  char i;

  sprite->animcount++;	// make the character's feet move, whether we can move or not
  
  // calculate sprite boundaries after move
  sp1_MakeRect8Pix(sprite->s, &r);
  x1 = pixtomaze(r.x+dx, mazerect.col);
  y1 = pixtomaze(r.y+dy, mazerect.row);
  x2 = pixtomaze(r.x+dx+15, mazerect.col);
  y2 = pixtomaze(r.y+dy+15, mazerect.row);
  
  // can we move there?
  if (checkforhouse(x1, x2, y1, y2)) {
      // no :(
      if(dx) {	// if we are trying to move horizontally...
        // can we move there if we throw away the y compound?
        yy1 = pixtomaze(r.y, mazerect.row);
        yy2 = pixtomaze(r.y+15, mazerect.row);
        if(!checkforhouse(x1, x2, yy1, yy2)) {
          // yes. do so.
          dy = 0; goto done;
        }
        
        // can we move there if we slide a little in y direction?
        // (this is important to avoid the player getting stuck on the edges of walls.)
        for(i=-4 ; i!= 5 ; i++)
        {
          if(!i) continue;	// not moving in y direction has been tried before.
          yy1 = pixtomaze(r.y+i, mazerect.row);
          yy2 = pixtomaze(r.y+15+i, mazerect.row);
          if(!checkforhouse(x1, x2, yy1, yy2))
          {
            // moving i pixels in y direction works; do so.
            dy+=i; goto done;
          }
        }
      }
      
      if(dy) {	// if we are trying to move vertically...
        // can we move there if we throw away the x compound?
        xx1 = pixtomaze(r.x, mazerect.col);
        xx2 = pixtomaze(r.x+15, mazerect.col);
        if(!checkforhouse(xx1, xx2, y1, y2)) {
          // yes. do so.
          dx = 0; goto done;
        }
        
        // can we move there if we slide a little in x direction?
        // (this is important to avoid the player getting stuck on the edges of walls.)
        for(i=-4; i!=5; i++)
        {
          if(!i) continue;	// not moving in x direction has been tried before.
          xx1 = pixtomaze(r.x+i, mazerect.col);
          xx2 = pixtomaze(r.x+15+i, mazerect.col);
          if(!checkforhouse(xx1, xx2, y1, y2))
          {
            // moving i pixels in x direction works; do so.
            dx+=i; goto done;
          }
        }
      }
      // tried everything to no avail. give up.
      dx=dy=0;
  }
done:
  sp1_MoveSprRel(sprite->s, &cr, (sprite->animcount&8)? anim : anim + (jibfront2- jibfront1), 0, 0, dy, dx);
}

// create a random maze.
void createmaze(void)
{
  int i,j,x;

  // clear the maze area.
  for(i=0; i!=15; i++) 
    for(j=0; j!=10; j++) settile(TILE_PFLASTER,INK_BLACK|PAPER_WHITE,i,j);

  // for debugging: setting int_seed to a constant value makes this predictable.
  // int_seed=1337;
  
  // create a frame of walls around the maze.
  for(i=0; i != 15; i++) {	// horizontal
    settile(TILE_HAUS,INK_RED|PAPER_MAGENTA,i,0);
    settile(TILE_HAUS,INK_RED|PAPER_MAGENTA,i,9);
  }
  for(i=0;i<10;i++) {		// vertical
    settile(TILE_HAUS,INK_RED|PAPER_MAGENTA,0,i);
    settile(TILE_HAUS,INK_RED|PAPER_MAGENTA,14,i);
  }

#if 1	// set to 0 to play in an empty maze.
  for(x=0;x<10*15*1;x++)
  {
    // pick a random spot in the maze
    i=rand()%6+2;
    j=rand()%11+2;

    // check for some conditions we want to avoid. doing so prevents inacessible
    // spaces from being created, among other things.
    
    // don't put corner on corner
    if(gettile(j-1,i-1)==TILE_HAUS && gettile(j,i-1)==TILE_PFLASTER && gettile(j-1,i)==TILE_PFLASTER) continue;
    if(gettile(j-1,i+1)==TILE_HAUS && gettile(j,i+1)==TILE_PFLASTER && gettile(j-1,i)==TILE_PFLASTER) continue;
    if(gettile(j+1,i-1)==TILE_HAUS && gettile(j,i-1)==TILE_PFLASTER && gettile(j+1,i)==TILE_PFLASTER) continue;
    if(gettile(j+1,i+1)==TILE_HAUS && gettile(j,i+1)==TILE_PFLASTER && gettile(j+1,i)==TILE_PFLASTER) continue;

    // don't put 2x2 blocks
    if(gettile(j-1,i-1)==TILE_HAUS && gettile(j-1,i)==TILE_HAUS && gettile(j,i-1)==TILE_HAUS) continue;
    if(gettile(j-1,i+1)==TILE_HAUS && gettile(j-1,i)==TILE_HAUS && gettile(j,i+1)==TILE_HAUS) continue;
    if(gettile(j+1,i-1)==TILE_HAUS && gettile(j+1,i)==TILE_HAUS && gettile(j,i-1)==TILE_HAUS) continue;
    if(gettile(j+1,i+1)==TILE_HAUS && gettile(j+1,i)==TILE_HAUS && gettile(j,i+1)==TILE_HAUS) continue;
    
    // don't close off spaces
    if(gettile(j,i-1)==TILE_HAUS && gettile(j,i+1)==TILE_HAUS) continue;
    if(gettile(j-1,i)==TILE_HAUS && gettile(j+1,i)==TILE_HAUS) continue;
    
    // everything ok, we can place the wall.
    settile(TILE_HAUS,INK_RED|PAPER_MAGENTA,j,i);
  }
#endif
  
}

uchar scorestring[10];	// score, converted to string for printing.

int main(void)
{
   uchar i,j;
   struct sp1_ss *s;		// temporary sprite pointer
   uchar jdx, jdy;		// player speed
   uchar joy;			// joystick state
   uchar* jan;			// pointer to player animation phase
   uchar xprize, yprize;	// new position of prizes
   
   void *heapend;		// pointer to free memory used in allocating balloc queues
   
   // set up interrupt mode 2.
   #asm
   di
   #endasm
   im2_Init(1, 0xd4d4);
   im2_CreateGenericISR(1, 0xd4d4);
   im2_RegHookFirst(255, my_isr);
   #asm
   ei
   #endasm
   
   hiscore = 0;	// reset hi-score.

restart:

   lives = 4;	// reset player lives
   score = 0;	// reset score
   
   // Initialize BALLOC.LIB
   ba_Init(2);	// three queues. reiniting the allocator saves us the trouble of freeing all allocated memory.
                // (calling ba_Init(3) causes corruption in data below(!) 0xc000. calling ba_Init(2) works fine,
                // even though we actually have three queues.)
   heapend = ba_AddMem(0, 10, sizeof(struct sp1_ss), 0xc000);	// one for sp1.lib
   heapend = ba_AddMem(1, 90, sizeof(struct sp1_cs), heapend);	// another one for sp1.lib
   ba_AddMem(2, 1, 20*30*sizeof(struct sp1_tp), heapend);	// one for the maze (only has a single element)
   maze = ba_Malloc(2);
   
   border(YELLOW);

   // initialize the sprite lib and clear the screen with spaces
   sp1_Initialize(SP1_IFLAG_MAKE_ROTTBL | SP1_IFLAG_OVERWRITE_TILES | SP1_IFLAG_OVERWRITE_DFILE, idtypetbl, sprdrawftbl, INK_BLACK | PAPER_YELLOW, ' ');

   // load the bg tile data for walls and floor (2x2 characters each)
   for(i=0; i!= 4; i++)
   {
     sp1_TileEntry(TILE_HAUS+i, haus+(i * 8));
     sp1_TileEntry(TILE_PFLASTER+i, pflaster);
   }

   // create the NPC sprite.
   s = uli.s = sp1_CreateSpr(sp1_SprId2Type(SP1_ID_MASK2_LB) | SP1_TYPE_2BYTE, 3, 0, 1);
   sp1_AddColSpr(s, sp1_SprId2Type(SP1_ID_MASK2), ulifront15-ulifront1, 1);
   sp1_AddColSpr(s, sp1_SprId2Type(SP1_ID_MASK2_RB), 0, 1);
   
   // create the player sprite.
   s = jib.s = sp1_CreateSpr(sp1_SprId2Type(SP1_ID_MASK2_LB) | SP1_TYPE_2BYTE, 3, 0, 1);
   sp1_AddColSpr(s, sp1_SprId2Type(SP1_ID_MASK2), ulifront15-ulifront1, 1);
   sp1_AddColSpr(s, sp1_SprId2Type(SP1_ID_MASK2_RB), 0, 1);

   // create the prizes
   for(i=0 ; i != 3 ; i++)
   {
     s = prizes[i].s = sp1_CreateSpr(sp1_SprId2Type(SP1_ID_MASK2_LB) | SP1_TYPE_2BYTE, 3, 0, 5);
     sp1_AddColSpr(s, sp1_SprId2Type(SP1_ID_MASK2), diamond15 - diamond1, 5);
     sp1_AddColSpr(s, sp1_SprId2Type(SP1_ID_MASK2_RB), 0, 5);
     sp1_IterateSprChar(s, colorSpr);
   }
   
   welcome();	// show the start screen
   
   // draw explanatory text and initial values
   my_printat(0,1,INK_BLUE|PAPER_YELLOW,"Lives");
   my_printat(0,24,INK_BLUE|PAPER_YELLOW,"Account");
   my_printat(0,11,INK_BLUE|PAPER_YELLOW,"Hi-Score");
   if(hiscore) {
     my_snprint(scorestring,hiscore,10);
     my_printat(1,10,INK_BLACK|PAPER_YELLOW,scorestring);
   } else {
     my_printat(1,10,INK_BLACK|PAPER_YELLOW,"        0");
   }
   my_printat(1,22,INK_BLACK|PAPER_WHITE,"        0");
   my_printat(23,8,INK_BLUE|PAPER_YELLOW,"Jib Goes Shopping");

   // create the maze and draw it.
   createmaze();
   sp1_PutTiles(&mazerect, maze);
   
   sp1_UpdateNow();
   
   resetsprites();	// initialize all sprite positions.
  
   // the game loop.
   while (1) {
      if(joytype != -2)
        joy = joystick(joytype);	// read joystick status.
      else {
        joy = 0;
        if(in_KeyPressed(keycodes[0])) joy += MOVE_UP;
        if(in_KeyPressed(keycodes[1])) joy += MOVE_DOWN;
        if(in_KeyPressed(keycodes[2])) joy += MOVE_LEFT;
        if(in_KeyPressed(keycodes[3])) joy += MOVE_RIGHT;
      }
      jdy = jdx = 0;		// reset player movement.
      jan = 0;			// reset player animation phase (redundant?)
      
      // move player, using the appropriate animation
#ifdef USE_EVEN
      if(joy & MOVE_DOWN) { jdy+=even?jib.dy_even:jib.dy_odd; jan = jibfront1; }
      if(joy & MOVE_UP) { jdy-=even?jib.dy_even:jib.dy_odd; jan = jibback1; }
      if(joy & MOVE_LEFT) { jdx-=even?jib.dx_even:jib.dx_odd; jan = jibleft1; }
      if(joy & MOVE_RIGHT) { jdx+=even?jib.dx_even:jib.dx_odd; jan = jibright1; }
#else
      if(joy & MOVE_DOWN) { jdy+=jib.dy_odd; jan = jibfront1; }
      if(joy & MOVE_UP) { jdy-=jib.dy_odd; jan = jibback1; }
      if(joy & MOVE_LEFT) { jdx-=jib.dx_odd; jan = jibleft1; }
      if(joy & MOVE_RIGHT) { jdx+=jib.dx_odd; jan = jibright1; }
#endif
      if(jdy||jdx) trymove(&jib, jdy, jdx, jan);
      
      // prizes
      for(i=0; i!= 3 ; i++)
      {
        if(prizes[i].offtime)	// prize not shown
        {
          prizes[i].offtime--;
          if(!prizes[i].offtime) // show it now
          {
            do {
              xprize = rand()%15;
              yprize = rand()%10;
            } while(gettile(xprize, yprize) != TILE_PFLASTER);
            sp1_MoveSprAbs(prizes[i].s,&cr,diamond1,yprize*2+2, xprize*2+1, 0, 0);
            prizes[i].ontime = rand()%512 + 256;	// random time for the prize to be shown
            continue;
          }
        }
        if(prizes[i].ontime)	// prize is shown
        {
          prizes[i].ontime--;
          if(!prizes[i].ontime)	// expired
          {
            sp1_MoveSprAbs(prizes[i].s,&cr,diamond1,0,32,0,0);	// move it off screen.
            prizes[i].offtime = rand()%512;	// random time until it reappears.
          }
        }
      }
      
      // check for a player/NPC collision (gotcha!)
      if(collide(jib.s,uli.s)) {
        lose_life();
        if(lives == 0) {
          game_over();
          goto restart;
        }
      }
      
      // check for a player/prize collision (kaching!)
      for(i=0; i!=3 ; i++) {
        if(collide(jib.s,prizes[i].s)) {
          score += 5;	// 500 points
          
          // reprint score
          my_snprint(scorestring,score, 10);
          my_printat(1,22,INK_RED|PAPER_WHITE, scorestring);
          
          // turn off prize
          sp1_MoveSprAbs(prizes[i].s,&cr,diamond1,0,32,0,0);
          prizes[i].ontime=0;
          prizes[i].offtime= rand()%512;
        }
      }
      
      if(moveuli())	// finished last move
        ulisnextstep();
        
      sp1_UpdateNow();  // draw screen now
      
      // wait for next frame to keep a constant game speed
      while(frames<1) {}
      frames=0;
      
#ifdef USE_EVEN
      if(even) even = 0;
      else even = 1;
#endif
      
   }

}

#asm

; black magic. sp1.lib does not work without this buffering data.
defb @11111111, @00000000
defb @11111111, @00000000
defb @11111111, @00000000
defb @11111111, @00000000
defb @11111111, @00000000
defb @11111111, @00000000
defb @11111111, @00000000

; sprite animation frames
#include "ulifront.asm"
#include "uliback.asm"
#include "ulileft.asm"
#include "uliright.asm"
#include "jibfront.asm"
#include "jibback.asm"
#include "jibleft.asm"
#include "jibright.asm"

; prize sprite frame
#include "diamond.asm"

; wall tiles
._haus
        DEFB    255,255,193,221,221,221,193,255
        DEFB    255,255,131,187,187,187,131,255
        DEFB    255,193,221,221,221,193,255,255
        DEFB    255,131,187,187,187,131,255,255

; floor tile
._pflaster
        DEFB     34, 34,255,136,136,136,255, 34

#endasm

#asm
; copied from z88dk. for some reason this does not end up in the spectrum libs.
._rand
   ld   hl,(_int_seed)
   ld   a,h
   add  a,a                     ;Set highest bit of seed to carry
   rl   l                       ;rotate L left (C<=L<=C)
   rl   h                       ;rotate H left (C<=L<=C)
   add  a,a                     ;Set second highest bit of seed to carry
   rl   l
   rl   h
   add  a,a
   rl   l
   rl   h
   ld   bc,$7415
   add  hl,bc                   ;Add $7415 to HL
   ld   (_int_seed),hl
   bit     7,h			;force to be +ve
   call    nz,l_neg
   ret
#endasm
